/*******************************************************************************
 * tlx/string/format_si_iec_units.hpp
 *
 * Part of tlx - http://panthema.net/tlx
 *
 * Copyright (C) 2016-2017 Timo Bingmann <tb@panthema.net>
 *
 * All rights reserved. Published under the Boost Software License, Version 1.0
 ******************************************************************************/

#ifndef TLX_STRING_FORMAT_SI_IEC_UNITS_HEADER
#define TLX_STRING_FORMAT_SI_IEC_UNITS_HEADER

#include <tlx/string/format_iec_units.hpp>
#include <tlx/string/format_si_units.hpp>

#endif // !TLX_STRING_FORMAT_SI_IEC_UNITS_HEADER

/******************************************************************************/
